"""Contains the class that represents the AWS Step Functions Fail State"""
from typing import Dict

from .base import State


class FailState(State):
    """Fail state

    A Fail state stops the state machine and surfaces error information. Fail states
    are generated by raising an exception in an .sfn file.

    See: https://docs.aws.amazon.com/step-functions/latest/dg/amazon-states-language-fail-state.html

    For example::

        raise Exception("oops")

    resolves to::

        {
            "Type": "Fail",
            "Error": "Exception",
            "Cause" "oops"
        }
    """

    TERMINAL = True

    def to_dict(self) -> Dict:
        """Return a serialized representation of the Fail state.

        The ast_node is an Exception object.
        """
        cause = (
            self.ast_node.exc.args[0].s
            if len(self.ast_node.exc.args) > 0
            else self.ast_node.exc.func.id
        )
        return {"Type": "Fail", "Error": self.ast_node.exc.func.id, "Cause": cause}
